import logging
import sys
from copy import deepcopy

import click
import click.shell_completion

from libvcs.shortcuts import create_project_from_pip_url

from ..config import filter_repos, find_config_files, load_configs

log = logging.getLogger(__name__)


def get_repo_completions(ctx: click.core.Context, args, incomplete):
    configs = (
        load_configs(find_config_files(include_home=True))
        if ctx.params["config"] is None
        else load_configs(files=[ctx.params["config"]])
    )
    found_repos = []
    repo_terms = [incomplete]

    for repo_term in repo_terms:
        dir, vcs_url, name = None, None, None
        if any(repo_term.startswith(n) for n in ["./", "/", "~", "$HOME"]):
            dir = repo_term
        elif any(repo_term.startswith(n) for n in ["http", "git", "svn", "hg"]):
            vcs_url = repo_term
        else:
            name = repo_term

        # collect the repos from the config files
        found_repos.extend(filter_repos(configs, dir=dir, vcs_url=vcs_url, name=name))
    if len(found_repos) == 0:
        found_repos = configs

    return [o["name"] for o in found_repos if o["name"].startswith(incomplete)]


def get_config_file_completions(ctx, args, incomplete):
    return [
        click.shell_completion.CompletionItem(c)
        for c in find_config_files(include_home=True)
        if str(c).startswith(incomplete)
    ]


def clamp(n, _min, _max):
    return max(_min, min(n, _max))


@click.command(name="sync")
@click.argument(
    "repo_terms", type=click.STRING, nargs=-1, shell_complete=get_repo_completions
)
@click.option(
    "config",
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="Specify config",
    shell_complete=get_config_file_completions,
)
def sync(repo_terms, config):
    if config:
        configs = load_configs([config])
    else:
        configs = load_configs(find_config_files(include_home=True))
    found_repos = []

    if repo_terms:
        for repo_term in repo_terms:
            dir, vcs_url, name = None, None, None
            if any(repo_term.startswith(n) for n in ["./", "/", "~", "$HOME"]):
                dir = repo_term
            elif any(repo_term.startswith(n) for n in ["http", "git", "svn", "hg"]):
                vcs_url = repo_term
            else:
                name = repo_term

            # collect the repos from the config files
            found_repos.extend(
                filter_repos(configs, dir=dir, vcs_url=vcs_url, name=name)
            )
    else:
        found_repos = configs

    list(map(update_repo, found_repos))


def progress_cb(output, timestamp):
    sys.stdout.write(output)
    sys.stdout.flush()


def update_repo(repo_dict):
    repo_dict = deepcopy(repo_dict)
    if "pip_url" not in repo_dict:
        repo_dict["pip_url"] = repo_dict.pop("url")
    repo_dict["progress_callback"] = progress_cb

    r = create_project_from_pip_url(**repo_dict)  # Creates the repo object
    r.update_repo(set_remotes=True)  # Creates repo if not exists and fetches

    return r



import pathlib

import pytest

from vcspull.util import get_config_dir


def test_vcspull_configdir_env_var(
    tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch
):
    monkeypatch.setenv("VCSPULL_CONFIGDIR", str(tmp_path))

    assert get_config_dir() == tmp_path


def test_vcspull_configdir_xdg_config_dir(
    tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch
):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    vcspull_dir = tmp_path / "vcspull"
    vcspull_dir.mkdir()

    assert get_config_dir() == vcspull_dir


def test_vcspull_configdir_no_xdg(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv("XDG_CONFIG_HOME")
    assert get_config_dir()



"""Tests for placing config dicts into :py:class:`Project` objects."""
import os

from _pytest.compat import LEGACY_PATH

from libvcs import BaseProject, GitProject, MercurialProject, SubversionProject
from libvcs.shortcuts import create_project_from_pip_url
from vcspull.config import filter_repos

from .fixtures import example as fixtures


def test_filter_dir():
    """`filter_repos` filter by dir"""
    repo_list = filter_repos(fixtures.config_dict_expanded, dir="*github_project*")

    assert len(repo_list) == 1
    for r in repo_list:
        assert r["name"] == "kaptan"


def test_filter_name():
    """`filter_repos` filter by name"""
    repo_list = filter_repos(fixtures.config_dict_expanded, name=".vim")

    assert len(repo_list) == 1
    for r in repo_list:
        assert r["name"] == ".vim"


def test_filter_vcs():
    """`filter_repos` filter by vcs remote url"""
    repo_list = filter_repos(fixtures.config_dict_expanded, vcs_url="*kernel.org*")

    assert len(repo_list) == 1
    for r in repo_list:
        assert r["name"] == "linux"


def test_to_dictlist():
    """`filter_repos` pulls the repos in dict format from the config."""
    repo_list = filter_repos(fixtures.config_dict_expanded)

    for r in repo_list:
        assert isinstance(r, dict)
        assert "name" in r
        assert "parent_dir" in r
        assert "url" in r

        if "remotes" in r:
            assert isinstance(r["remotes"], list)
            for remote in r["remotes"]:
                assert isinstance(remote, dict)
                assert "remote_name" == remote
                assert "url" == remote


def test_vcs_url_scheme_to_object(tmpdir: LEGACY_PATH):
    """Verify `url` return {Git,Mercurial,Subversion}Project.

    :class:`GitProject`, :class:`MercurialProject` or :class:`SubversionProject`
    object based on the pip-style URL scheme.

    """
    git_repo = create_project_from_pip_url(
        **{
            "pip_url": "git+git://git.myproject.org/MyProject.git@da39a3ee5e6b4b",
            "dir": str(tmpdir.join("myproject1")),
        }
    )

    # TODO cwd and name if duplicated should give an error

    assert isinstance(git_repo, GitProject)
    assert isinstance(git_repo, BaseProject)

    hg_repo = create_project_from_pip_url(
        **{
            "pip_url": "hg+https://hg.myproject.org/MyProject#egg=MyProject",
            "dir": str(tmpdir.join("myproject2")),
        }
    )

    assert isinstance(hg_repo, MercurialProject)
    assert isinstance(hg_repo, BaseProject)

    svn_repo = create_project_from_pip_url(
        **{
            "pip_url": "svn+svn://svn.myproject.org/svn/MyProject#egg=MyProject",
            "dir": str(tmpdir.join("myproject3")),
        }
    )

    assert isinstance(svn_repo, SubversionProject)
    assert isinstance(svn_repo, BaseProject)


def test_to_repo_objects(tmpdir: LEGACY_PATH):
    """:py:obj:`dict` objects into Project objects."""
    repo_list = filter_repos(fixtures.config_dict_expanded)
    for repo_dict in repo_list:
        r = create_project_from_pip_url(**repo_dict)

        assert isinstance(r, BaseProject)
        assert r.name
        assert r.name == repo_dict["name"]
        assert r.parent_dir
        assert r.parent_dir == repo_dict["parent_dir"]
        assert r.url
        assert r.url == repo_dict["url"]

        assert r.path == os.path.join(r.parent_dir, r.name)

        if "remotes" in repo_dict:
            assert isinstance(r.remotes, list)
            for remote_name, remote_dict in r.remotes.items():
                assert isinstance(remote_dict, dict)
                assert "fetch_url" in remote_dict
                assert "push_url" in remote_dict



class VCSPullException(Exception):

    """Standard exception raised by vcspull."""


class MultipleConfigWarning(VCSPullException):
    message = "Multiple configs found in home directory use only one." " .yaml, .json."



"""CLI utilities for vcspull.

vcspull.cli
~~~~~~~~~~~

"""
import logging

import click

from libvcs.__about__ import __version__ as libvcs_version

from ..__about__ import __version__
from ..log import setup_logger
from .sync import sync

log = logging.getLogger(__name__)


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
        "allow_interspersed_args": True,
    }
)
@click.option(
    "--log-level",
    default="INFO",
    help="Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
)
@click.version_option(
    __version__,
    "-V",
    "--version",
    message=f"%(prog)s %(version)s, libvcs {libvcs_version}",
)
def cli(log_level):
    setup_logger(log=log, level=log_level.upper())


# Register sub-commands here
cli.add_command(sync)



"""Helpers for vcspull."""
import os
import pathlib
import textwrap
from typing import Literal

import kaptan


class EnvironmentVarGuard:

    """Class to help protect the environment variable properly.

    May be used as context manager.
    Vendorize to fix issue with Anaconda Python 2 not
    including test module, see #121.
    """

    def __init__(self):
        self._environ = os.environ
        self._unset = set()
        self._reset = dict()

    def set(self, envvar, value):
        if envvar not in self._environ:
            self._unset.add(envvar)
        else:
            self._reset[envvar] = self._environ[envvar]
        self._environ[envvar] = value

    def unset(self, envvar):
        if envvar in self._environ:
            self._reset[envvar] = self._environ[envvar]
            del self._environ[envvar]

    def __enter__(self):
        return self

    def __exit__(self, *ignore_exc):
        for envvar, value in self._reset.items():
            self._environ[envvar] = value
        for unset in self._unset:
            del self._environ[unset]


def write_config(config_path: pathlib.Path, content: str) -> pathlib.Path:
    config_path.write_text(content, encoding="utf-8")
    return config_path


def import_raw(data: str, format: Literal["yaml", "json"]) -> kaptan.Kaptan:
    return kaptan.Kaptan(handler=format).import_config(textwrap.dedent(data))


def load_raw(data: str, format: Literal["yaml", "json"]) -> dict:
    return import_raw(data=data, format=format).export("dict")
